import functools
import json
import logging
from logging.config import fileConfig
from types import FunctionType
from typing import Dict, List, Optional, Union

from pkg_resources import resource_filename

from magnus import defaults, exceptions, graph, nodes, pipeline, utils
from magnus.interaction import get_tag

logger = logging.getLogger(defaults.NAME)


# class step(object):

#     def __init__(
#             self, name: Union[str, FunctionType],
#             catalog_config: dict = None, magnus_config: str = None,
#             parameters_file: str = None):
#         """
#         This decorator could be used to make the function within the scope of magnus.

#         Since we are not orchestrating, it is expected that resource management happens outside this scope.

#         Args:
#             name (str, callable): The name of the step. The step log would have the same name
#             catalog_config (dict): The configuration of the catalog per step.
#             magnus_config (str): The name of the file having the magnus config, defaults to None.
#         """
#         if isinstance(name, FunctionType):
#             name = name()

#         self.name = name
#         self.catalog_config = catalog_config
#         self.active = True  # Check if we are executing the function via pipeline

#         if pipeline.global_executor \
#                 and pipeline.global_executor.execution_plan == defaults.EXECUTION_PLAN.CHAINED.value:
#             self.active = False
#             return

#         self.executor = pipeline.prepare_configurations(
#             configuration_file=magnus_config, parameters_file=parameters_file)

#         self.executor.execution_plan = defaults.EXECUTION_PLAN.UNCHAINED.value
#         run_id = self.executor.step_decorator_run_id
#         if not run_id:
#             msg = (
#                 f'Step decorator expects run id from environment.'
#             )
#             raise Exception(msg)

#         self.executor.run_id = run_id
#         utils.set_magnus_environment_variables(run_id=run_id, configuration_file=magnus_config, tag=get_tag())

#         try:
#             # Try to get it if previous steps have created it
#             # TODO: Can call the set_up_runlog now.
#             run_log = self.executor.run_log_store.get_run_log_by_id(self.executor.run_id)
#             if run_log.status in [defaults.FAIL, defaults.SUCCESS]:  # TODO: Remove this in preference to defaults
#                 """
#                 This check is mostly useless as we do not know when the graph ends as they are created dynamically.
#                 This only prevents from using a run_id which has reached a final state.
#                 #TODO: There is a need to create a status called step_success
#                 """
#                 msg = (
#                     f'The run_log for run_id: {run_id} already exists and is in {run_log.status} state.'
#                     ' Make sure that this was not run before.'
#                 )
#                 raise Exception(msg)
#         except exceptions.RunLogNotFoundError:
#             # Create one if they are not created
#             self.executor._set_up_run_log()

#     def __call__(self, func):
#         """
#         The function is converted into a node and called via the magnus framework.
#         """
#         @functools.wraps(func)
#         def wrapped_f(*args, **kwargs):
#             if not self.active:
#                 # If we are not running via decorator, execute the function
#                 return func(*args, **kwargs)

#             step_config = {
#                 'command': func,
#                 'command_type': 'python-function',
#                 'type': 'task',
#                 'next': 'not defined',
#                 'catalog': self.catalog_config
#             }
#             node = graph.create_node(name=self.name, step_config=step_config)
#             self.executor.execute_from_graph(node=node)
#             run_log = self.executor.run_log_store.get_run_log_by_id(run_id=self.executor.run_id, full=False)
#             # TODO: If the previous step succeeded, make the status of the run log step_success
#             print(json.dumps(run_log.dict(), indent=4))
#         return wrapped_f


class Task:
    def __init__(self, name: str, command: Union[str, FunctionType], command_type: str = defaults.COMMAND_TYPE,
                 command_config: Optional[dict] = None, catalog: Optional[dict] = None,
                 mode_config: Optional[dict] = None, retry: int = 1, on_failure: str = '',
                 next_node: str = ''):
        self.name = name
        self.command = command
        self.command_type = command_type
        self.command_config = command_config or {}
        self.catalog = catalog or {}
        self.mode_config = mode_config or {}
        self.retry = retry
        self.on_failure = on_failure
        self.next_node = next_node or "success"
        self.node: Optional[nodes.BaseNode] = None

    def _construct_node(self):
        # TODO: The below has issues if the function and the pipeline are in the same module
        # Something to do with __main__ being present
        if isinstance(self.command, FunctionType):
            self.command = utils.get_module_and_func_from_function(self.command)

        node_config = {
            'type': 'task',
            'next_node': self.next_node,
            'command': self.command,
            'command_type': self.command_type,
            'command_config': self.command_config,
            'catalog': self.catalog,
            'mode_config': self.mode_config,
            'retry': self.retry,
            'on_failure': self.on_failure
        }
        # The node will temporarily have invalid branch names
        self.node = graph.create_node(name=self.name, step_config=node_config, internal_branch_name='')

    def _fix_internal_name(self):
        pass


class AsIs:
    def __init__(self, name: str, mode_config: Optional[dict] = None, retry: int = 1, on_failure: str = '',
                 next_node: str = '', **kwargs):
        self.name = name
        self.mode_config = mode_config or {}
        self.retry = retry
        self.on_failure = on_failure
        self.next_node = next_node or "success"
        self.additional_kwargs = kwargs or {}
        self.node: Optional[nodes.BaseNode] = None

    def _construct_node(self):
        node_config = {
            'type': 'as-is',
            'next_node': self.next_node,
            'mode_config': self.mode_config,
            'retry': self.retry,
            'on_failure': self.on_failure
        }
        node_config.update(self.additional_kwargs)
        # The node will temporarily have invalid branch names
        self.node = graph.create_node(name=self.name, step_config=node_config, internal_branch_name='')

    def _fix_internal_name(self):
        pass


class Pipeline:
    # A way for the user to define a pipeline
    # TODO: Allow for nodes other than Task, AsIs
    def __init__(
            self, start_at: Union[Task, AsIs],
            name: str = '', description: str = '', max_time: int = defaults.MAX_TIME, internal_branch_name: str = ''):
        self.start_at = start_at
        self.name = name
        self.description = description
        self.max_time = max_time
        self.internal_branch_name = internal_branch_name
        self.dag: Optional[graph.Graph] = None

    def construct(self, steps: List[Task]):
        graph_config: Dict[str, Union[str, int]] = {
            'description': self.description,
            'name': self.name,
            'max_time': self.max_time,
            'internal_branch_name': self.internal_branch_name
        }
        messages: List[str] = []
        for step in steps:
            step._construct_node()
            messages.extend(step.node.validate())  # type: ignore

        if not steps:
            raise Exception('A dag needs at least one step')

        if messages:
            raise Exception(', '.join(messages))

        graph_config['start_at'] = self.start_at.node.name  # type: ignore

        dag = graph.Graph(**graph_config)  # type: ignore
        dag.nodes = [step.node for step in steps]  # type: ignore

        dag.add_terminal_nodes()

        dag.validate()
        self.dag = dag

    def execute(self, configuration_file: str = '', run_id: str = '', tag: str = '', parameters_file: str = '',
                log_level: str = defaults.LOG_LEVEL):

        fileConfig(resource_filename(__name__, 'log_config.ini'))
        logger = logging.getLogger(defaults.NAME)
        logger.setLevel(log_level)

        run_id = utils.generate_run_id(run_id=run_id)
        mode_executor = pipeline.prepare_configurations(
            configuration_file=configuration_file,
            run_id=run_id,
            tag=tag,
            parameters_file=parameters_file)

        mode_executor.execution_plan = defaults.EXECUTION_PLAN.CHAINED.value
        utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

        mode_executor.dag = self.dag
        # Prepare for graph execution
        mode_executor.prepare_for_graph_execution()

        logger.info('Executing the graph')
        mode_executor.execute_graph(dag=mode_executor.dag)

        mode_executor.send_return_code()



import pickle
from typing import Any


class BasePickler:

    extension = ''
    service_name = ''

    def dump(self, data: Any, path: str):
        """
        Dump an object to the specified path.
        The path is the full path.

        To correctly identify the pickler from possible implementations, we use the extension.
        An extension is added automatically, if not provided.

        Args:
            data (Any): The object to pickle
            path (str): The path to save the pickle file

        Raises:
            NotImplementedError: Base class has no implementation
        """
        raise NotImplementedError

    def load(self, path: str) -> Any:
        """
        Load the object from the specified path.

        To correctly identify the pickler from possible implementations, we use the extension.
        An extension is added automatically, if not provided.

        Args:
            path (str): The path to load the pickled file from.

        Raises:
            NotImplementedError: Base class has no implementation.
        """
        raise NotImplementedError


class NativePickler(BasePickler):
    """
    Uses native python pickle to load and dump files
    """

    extension = ".pickle"
    service_name = "pickle"

    def dump(self, data: Any, path: str):
        if not path.endswith(self.extension):
            path = path + self.extension

        with open(path, 'wb') as f:
            pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)

    def load(self, path: str) -> Any:
        if not path.endswith(self.extension):
            path = path + self.extension

        data = None
        with open(path, 'rb') as f:
            data = pickle.load(f)

        return data



import contextlib
import importlib
import io
import json
import logging
import os
import subprocess
import sys

from pydantic import BaseModel

from magnus import defaults, put_in_catalog, utils

logger = logging.getLogger(defaults.NAME)

try:
    import ploomber_engine as pm
except ImportError:
    pm = None


@contextlib.contextmanager
def output_to_file(path: str):
    """
    Context manager to put the output of a function execution to catalog

    Args:
        path (str): Mostly the command you are executing.

    """
    log_file = open(f"{path}.log", 'w')
    f = io.StringIO()
    try:
        with contextlib.redirect_stdout(f):
            yield
    finally:
        print(f.getvalue())  # print to console
        log_file.write(f.getvalue())  # Print to file

        f.close()
        log_file.close()
        put_in_catalog(log_file.name)
        os.remove(log_file.name)


class BaseTaskType:  # pylint: disable=too-few-public-methods
    """
    A base task class which does the execution of command defined by the user
    """
    task_type = ''

    class Config(BaseModel):
        command: str

    def __init__(self, config: dict = None):
        config = config or {}
        self.config = self.Config(**config)

    @property
    def command(self):
        return self.config.command

    def _to_dict(self) -> dict:
        """
        Return a dictionary representation of the task
        """
        task = {}
        task['command'] = self.command
        task['config'] = self.config

        return task

    def _get_parameters(self, map_variable: dict = None, **kwargs) -> dict:
        """
        Return the parameters in scope for the execution

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        Returns:
            dict: The parameters dictionary in-scope for the task execution
        """
        return utils.get_user_set_parameters(remove=False)

    def execute_command(self, map_variable: dict = None, **kwargs):
        """
        The function to execute the command.

        And map_variable is sent in as an argument into the function.

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        Raises:
            NotImplementedError: Base class, not implemented
        """
        raise NotImplementedError()

    def _set_parameters(self, parameters: dict = None, **kwargs):
        """
        Set the parameters back to the environment variables.

        Args:
            parameters (dict, optional): The parameters to set back as env variables. Defaults to None.
        """
        # Nothing to do
        if not parameters:
            return

        if not isinstance(parameters, dict):
            msg = (
                f'call to function {self.command} returns of type: {type(parameters)}. '
                'Only dictionaries are supported as return values for functions as part part of magnus pipeline.')
            logger.warn(msg)
            return

        for key, value in parameters.items():
            logger.info(f'Setting User defined parameter {key} with value: {value}')
            os.environ[defaults.PARAMETER_PREFIX + key] = json.dumps(value)


class PythonFunctionType(BaseTaskType):
    task_type = 'python-function'

    def execute_command(self, map_variable: dict = None, **kwargs):
        parameters = self._get_parameters()
        filtered_parameters = utils.filter_arguments_for_func(self.command, parameters, map_variable)

        if map_variable:
            os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE'] = json.dumps(map_variable)

        logger.info(f'Calling {self.command} with {filtered_parameters}')

        with output_to_file(self.command) as _:
            try:
                user_set_parameters = self.command(**filtered_parameters)
            except Exception as _e:
                msg = (
                    f'Call to the function {self.command} with {filtered_parameters} did not succeed.\n'
                )
                logger.exception(msg)
                logger.exception(_e)
                raise

            if map_variable:
                del os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE']

            self._set_parameters(user_set_parameters)


class PythonTaskType(BaseTaskType):  # pylint: disable=too-few-public-methods
    """
    The task class for python command
    """
    task_type = 'python'

    def execute_command(self, map_variable: dict = None, **kwargs):
        module, func = utils.get_module_and_func_names(self.command)
        sys.path.insert(0, os.getcwd())  # Need to add the current directory to path
        imported_module = importlib.import_module(module)
        f = getattr(imported_module, func)

        parameters = self._get_parameters()
        filtered_parameters = utils.filter_arguments_for_func(f, parameters, map_variable)

        if map_variable:
            os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE'] = json.dumps(map_variable)

        logger.info(f'Calling {func} from {module} with {filtered_parameters}')
        with output_to_file(self.command) as _:
            try:
                user_set_parameters = f(**filtered_parameters)
            except Exception as _e:
                msg = (
                    f'Call to the function {self.command} with {filtered_parameters} did not succeed.\n'
                )
                logger.exception(msg)
                logger.exception(_e)
                raise

            if map_variable:
                del os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE']

            self._set_parameters(user_set_parameters)


class PythonLambdaTaskType(BaseTaskType):  # pylint: disable=too-few-public-methods
    """
    The task class for python-lambda command
    """
    task_type = 'python-lambda'

    def execute_command(self, map_variable: dict = None, **kwargs):
        if '_' in self.command or '__' in self.command:
            msg = (
                f'Command given to {self.task_type} cannot have _ or __ in them. '
                'The string is supposed to be for simple expressions only.'
            )
            raise Exception(msg)

        f = eval(self.command)

        parameters = self._get_parameters()
        filtered_parameters = utils.filter_arguments_for_func(f, parameters, map_variable)

        if map_variable:
            os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE'] = json.dumps(map_variable)

        logger.info(f'Calling lambda function: {self.command} with {filtered_parameters}')
        try:
            user_set_parameters = f(**filtered_parameters)
        except Exception as _e:
            msg = (
                f'Call to the function {self.command} with {filtered_parameters} did not succeed.\n'
            )
            logger.exception(msg)
            logger.exception(_e)
            raise

        if map_variable:
            del os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE']

        self._set_parameters(user_set_parameters)


class NotebookTaskType(BaseTaskType):
    """
    The task class for Notebook based execution
    """
    task_type = 'notebook'

    class Config(BaseTaskType.Config):
        notebook_output_path: str = ''
        optional_ploomber_args: dict = {}

    @property
    def notebook_output_path(self):
        if self.config.notebook_output_path:
            return self.config.notebook_output_path

        return ''.join(self.command.split('.')[:-1]) + '_out.ipynb'

    def __init__(self, config: dict = None):
        super().__init__(config)

        if not self.config.command.endswith('.ipynb'):
            raise Exception('Notebook task should point to a ipynb file')

    def execute_command(self, map_variable: dict = None, **kwargs):
        try:
            if not pm:
                raise ImportError('Ploomber engine is required for notebook type node')

            parameters = self._get_parameters()
            filtered_parameters = parameters

            if map_variable:
                os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE'] = json.dumps(map_variable)

            notebook_output_path = self.notebook_output_path

            ploomber_optional_args = self.config.optional_ploomber_args  # type: ignore

            kwds = {
                'input_path': self.command,
                'output_path': notebook_output_path,
                'parameters': filtered_parameters,
                'log_output': True,
                'progress_bar': False
            }

            kwds.update(ploomber_optional_args)

            pm.execute_notebook(**kwds)

            put_in_catalog(notebook_output_path)
            if map_variable:
                del os.environ[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE']

        except ImportError as e:
            msg = (
                f'Task type of notebook requires ploomber engine to be installed. Please install via optional: notebook'
            )
            raise Exception(msg) from e


class ShellTaskType(BaseTaskType):
    """
    The task class for shell based commands
    """
    task_type = 'shell'

    def execute_command(self, map_variable: dict = None, **kwargs):
        # TODO can we do this without shell=True. Hate that but could not find a way out
        # This is horribly weird, focussing only on python ways of doing for now
        # It might be that we have to write a bash/windows script that does things for us
        # Need to over-ride set parameters too
        subprocess_env = os.environ.copy()

        if map_variable:
            subprocess_env[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE'] = json.dumps(map_variable)

        result = subprocess.run(self.command, check=True, env=subprocess_env, shell=True,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        print(result.stdout)
        print(result.stderr)



from __future__ import annotations

import json
import logging
import os
from pathlib import Path
from typing import Any, Union

from magnus import defaults, exceptions, pickler, pipeline, utils

logger = logging.getLogger(defaults.NAME)


def track_this(step: int = 0, **kwargs):
    """
    Set up the keyword args as environment variables for tracking purposes as
    part of the run.

    For every key-value pair found in kwargs, we set up an environmental variable of
    MAGNUS_TRACK_key_{step} = json.dumps(value)

    If step=0, we ignore step for magnus purposes.

    Args:
        kwargs (dict): The dictionary of key value pairs to track.
    """
    from magnus import context  # pylint: disable=import-outside-toplevel

    if not context.executor:
        msg = (
            f"There are no active executor and services. This should not have happened and is a bug."
            " Please raise a bug report."
        )
        raise Exception(msg)

    prefix = defaults.TRACK_PREFIX

    if step:
        prefix += f'{str(step)}_'

    for key, value in kwargs.items():
        logger.info(f'Tracking {key} with value: {value}')
        os.environ[prefix + key] = json.dumps(value)
        context.executor.experiment_tracker.set_metric(key, value, step=step)  # type: ignore


def store_parameter(update: bool = True, **kwargs: dict):
    """
    Set up the keyword args as environment variables for parameters tracking
    purposes as part pf the run.

    If update_existing is True, we override the current value if the parameter already exists.

    For every key-value pair found in kwargs, we set up an environmental variable of
    MAGNUS_PRM_key = json.dumps(value)
    """
    for key, value in kwargs.items():
        logger.info(f'Storing parameter {key} with value: {value}')
        environ_key = defaults.PARAMETER_PREFIX + key

        if environ_key in os.environ and not update:
            continue

        os.environ[environ_key] = json.dumps(value)


def get_parameter(key=None) -> Union[str, dict]:
    """
    Get the parameter set as part of the user convenience function.

    We do not remove the parameter from the environment in this phase as
    as the function execution has not been completed.

    Returns all the parameters, if no key was provided.

    Args:
        key (str, optional): The parameter key to retrieve. Defaults to None.

    Raises:
        Exception: If the mentioned key was not part of the paramters

    Returns:
        Union[str, dict]: A single value of str or a dictionary if no key was specified
    """
    parameters = utils.get_user_set_parameters(remove=False)
    if not key:
        return parameters
    if key not in parameters:
        raise Exception(f'Parameter {key} is not set before')
    return parameters[key]


def get_secret(secret_name: str = None) -> str:
    """
    Get a secret by the name from the secrets manager

    Args:
        secret_name (str): The name of the secret to get. Defaults to None.

    Returns:
        str: The secret from the secrets manager, if exists. If the requested secret was None, we return all.
        Otherwise, raises exception.

    Raises:
        exceptions.SecretNotFoundError: Secret not found in the secrets manager.
    """
    from magnus import context  # pylint: disable=import-outside-toplevel

    if not context.executor:
        msg = (
            f"There are no active executor and services. This should not have happened and is a bug."
            " Please raise a bug report."
        )
        raise Exception(msg)

    secrets_handler = context.executor.secrets_handler  # type: ignore
    try:
        return secrets_handler.get(name=secret_name)  # type: ignore
    except exceptions.SecretNotFoundError:
        logger.exception(f'No secret by the name {secret_name} found in the store')
        raise


def get_from_catalog(name: str, destination_folder: str = None):
    """
    A convenience interaction function to get file from the catalog and place it in the destination folder

    Note: We do not perform any kind of serialization/deserialization in this way.
    Args:
        name (str): The name of the file to get from the catalog
        destination_folder (None): The place to put the file. defaults to compute data folder

    """
    from magnus import context  # pylint: disable=import-outside-toplevel

    if not context.executor:
        msg = (
            f"There are no active executor and services. This should not have happened and is a bug."
            " Please raise a bug report."
        )
        raise Exception(msg)

    if not destination_folder:
        destination_folder = context.executor.catalog_handler.compute_data_folder  # type: ignore

    data_catalog = context.executor.catalog_handler.get(name, run_id=context.executor.run_id,  # type: ignore
                                                        compute_data_folder=destination_folder)

    if not data_catalog:
        logger.warn(f"No catalog was obtained by the {name}")

    if context.executor.context_step_log:  # type: ignore
        context.executor.context_step_log.add_data_catalogs(data_catalog)  # type: ignore
    else:
        logger.warning("Step log context was not found during interaction! The step log will miss the record")


def put_in_catalog(filepath: str):
    """
    A convenience interaction function to put the file in the catalog.

    Note: We do not perform any kind of serialization/deserialization in this way.

    Args:
        filepath (str): The path of the file to put in the catalog
    """
    from magnus import context  # pylint: disable=import-outside-toplevel

    if not context.executor:
        msg = (
            f"There are no active executor and services. This should not have happened and is a bug."
            " Please raise a bug report."
        )
        raise Exception(msg)

    file_path = Path(filepath)

    data_catalog = context.executor.catalog_handler.put(file_path.name, run_id=context.executor.run_id,  # type: ignore
                                                        compute_data_folder=file_path.parent)
    if not data_catalog:
        logger.warn(f"No catalog was done by the {filepath}")

    if context.executor.context_step_log:  # type: ignore
        context.executor.context_step_log.add_data_catalogs(data_catalog)  # type: ignore
    else:
        logger.warning("Step log context was not found during interaction! The step log will miss the record")


def put_object(data: Any, name: str):
    """
    A convenient interaction function to serialize and store the object in catalog.

    Args:
        data (Any): The data object to add to catalog
        name (str): The name to give to the object
    """
    native_pickler = pickler.NativePickler()

    native_pickler.dump(data=data, path=name)
    put_in_catalog(f"{name}{native_pickler.extension}")

    # Remove the file
    os.remove(f"{name}{native_pickler.extension}")


def get_object(name: str) -> Any:
    """
    A convenient interaction function to deserialize and retrieve the object from the catalog.

    Args:
        name (str): The name of the object to retrieve

    Returns:
        Any : The object
    """
    native_pickler = pickler.NativePickler()

    get_from_catalog(name=f"{name}{native_pickler.extension}", destination_folder='.')

    try:
        data = native_pickler.load(name)

        # Remove the file
        os.remove(f"{name}{native_pickler.extension}")
        return data
    except FileNotFoundError as e:
        msg = (
            f"No object by the name {name} has been put in the catalog before."
        )
        logger.exception(msg)
        raise e


def get_run_id() -> str:
    """
    Returns the run_id of the current run
    """
    return os.environ.get(defaults.ENV_RUN_ID, '')


def get_tag() -> str:
    """
    Returns the tag from the environment.

    Returns:
        str: The tag if provided for the run, otherwise None
    """
    return os.environ.get(defaults.MAGNUS_RUN_TAG, '')


def get_experiment_tracker_context():
    """
    Return a context session of the experiment tracker.

    You can start to use the context with the python with statement.

    eg:
    with get_experiment_tracker_context() as ctx:
        pass

    Returns:
        _type_: _description_
    """
    from magnus import context  # pylint: disable=import-outside-toplevel

    if not context.executor:
        msg = (
            f"There are no active executor and services. This should not have happened and is a bug."
            " Please raise a bug report."
        )
        raise Exception(msg)

    experiment_tracker = context.executor.experiment_tracker
    return experiment_tracker.client_context


def start_interactive_session(run_id: str = "", config_file: str = "", tag: str = "", parameters_file: str = ""):
    """
    During interactive python coding, either via notebooks or ipython, you can start a magnus session by calling
    this function. The executor would always be local executor as its interactive.

    If this was called during a pipeline/function/notebook execution, it will be ignored.

    Args:
        run_id (str, optional): The run id to use. Defaults to "" and would be created if not provided.
        config_file (str, optional): The configuration file to use. Defaults to "" and magnus defaults.
        tag (str, optional): The tag to attach to the run. Defaults to "".
        parameters_file (str, optional): The parameters file to use. Defaults to "".
    """
    from magnus import context  # pylint: disable=import-outside-toplevel
    from magnus import graph

    if context.executor:
        logger.warn("This is not an interactive session or a session has already been activated.")
        return

    run_id = utils.generate_run_id(run_id=run_id)
    executor = pipeline.prepare_configurations(configuration_file=config_file,
                                               run_id=run_id, tag=tag, parameters_file=parameters_file,
                                               force_local_executor=True)

    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=config_file, tag=tag)

    executor.execution_plan = defaults.EXECUTION_PLAN.INTERACTIVE.value
    executor.prepare_for_graph_execution()
    step_config = {
        'command': "interactive",
        'command_type': "python",
        'type': 'task',
        'next': 'success'
    }

    node = graph.create_node(name="interactive", step_config=step_config)
    step_log = executor.run_log_store.create_step_log("interactive", node._get_step_log_name())
    executor.add_code_identities(node=node, step_log=step_log)

    step_log.step_type = node.node_type
    step_log.status = defaults.PROCESSING
    executor.context_step_log = step_log


def end_interactive_session():
    from magnus import context  # pylint: disable=import-outside-toplevel

    if not context.executor:
        logger.warn("There is no active session in play, doing nothing!")
        return

    if context.executor.execution_plan != defaults.EXECUTION_PLAN.INTERACTIVE.value:
        logger.warn("There is not an interactive session, doing nothing!")
        return

    tracked_data = utils.get_tracked_data()
    parameters = utils.get_user_set_parameters(remove=True)

    step_log = context.executor.context_step_log
    step_log.user_defined_metrics = tracked_data
    context.executor.run_log_store.add_step_log(step_log, context.executor.run_id)

    context.executor.run_log_store.set_parameters(context.executor.run_id, parameters)

    context.executor.context_step_log = None
    context.executor.execution_plan = ""
    context.executor = None



import contextlib
import functools
import logging
from typing import Any

from pydantic import BaseModel

from magnus import defaults

logger = logging.getLogger(defaults.NAME)


class BaseExperimentTracker:
    """
    Base Experiment tracker class definition.
    """

    service_name = ''

    class Config(BaseModel):
        pass

    def __init__(self, config: dict = None, **kwargs):  # pylint: disable=unused-argument
        config = config or {}
        self.config = self.Config(**config)

    @property
    def client_context(self):
        return contextlib.nullcontext()

    def log_metric(self, key: str, value: float, step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (float): The value of the metric
            step (int): Optional step at which it was recorded

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        raise NotImplementedError

    def log_parameter(self, key: str, value: Any):
        """
        Logs a parameter in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (any): The value of the metric

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        pass


class DoNothingTracker(BaseExperimentTracker):
    """
    A Do nothing tracker
    """
    service_name = 'do-nothing'

    def set_metric(self, key: str, value: float, step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (float): The value of the metric
        """
        pass

    def log_parameter(self, key: str, value: Any):
        pass


# class MLFlowExperimentTracker(BaseExperimentTracker):
#     """
#     A MLFlow experiment tracker.

#     TODO: Need to set up credentials from secrets
#     """
#     service_name = 'mlflow'

#     class Config(BaseExperimentTracker.Config):
#         server_url: str
#         autolog: bool = False

#     def __init__(self, config, **kwargs):
#         super().__init__(config, **kwargs)
#         self.default_experiment_name = 'Default'
#         self.active_run_id = None

#         mlflow.set_tracking_uri(self.config.server_url)

#         if self.config.autolog:
#             mlflow.autolog(log_models=False)

#     @functools.cached_property
#     def experiment_id(self):
#         from magnus.pipeline import global_executor
#         experiment_name = self.default_experiment_name

#         # If a tag is provided, we should create that as our experiment
#         if global_executor.tag:
#             experiment_name = global_executor.tag

#         experiment = mlflow.get_experiment_by_name(experiment_name)
#         if not experiment:
#             # Create the experiment and get it.
#             experiment = mlflow.create_experiment(experiment_name)
#             experiment = mlflow.get_experiment(experiment)

#         return experiment.experiment_id

#     @functools.cached_property
#     def run_name(self):
#         from magnus.pipeline import global_executor
#         return global_executor.run_id

#     @property
#     def client_context(self):
#         if self.active_run_id:
#             return mlflow.start_run(run_id=self.active_run_id,
#                                     experiment_id=self.experiment_id, run_name=self.run_name)

#         active_run = mlflow.start_run(run_name=self.run_name, experiment_id=self.experiment_id)
#         self.active_run_id = active_run.info.run_id
#         return active_run

#     def set_metric(self, key: str, value: float, step: int = 0):
#         """
#         Sets the metric in the experiment tracking.

#         Args:
#             key (str): The key against you want to store the value
#             value (Any): The value of the metric
#         """
#         if not isinstance(value, float):
#             msg = (
#                 f'Only float values are accepted as metrics. Setting the metric {key} as parameter {key}_{step}'
#             )
#             logger.warning(msg)
#             self.log_parameter(key=key+f'_{step}', value=value)
#             return

#         with self.client_context as _:
#             mlflow.log_metric(key, value, step=step or None)

#     def get_metric(self, key: str) -> Any:
#         """
#         Return the metric by the key

#         Args:
#             key (str): The metric you want to retrieve

#         Returns:
#             Any: The value of the key
#         """
#         with self.client_context as client:
#             metrics = client.data.metrics
#             return metrics.get(key)

#     def log_parameter(self, key: str, value: Any):
#         with self.client_context as _:
#             mlflow.log_param(key, value)

#     def get_parameter(self, key: str) -> Any:
#         pass
